interface Bram#(type addr_t, type data_t);
    method Action readReq(addr_t addr);
    method data_t readResp();
    method Action write(addr_t addr, data_t data);
    method Action zero();
endinterface

import "BVI" bram_fu =
module mkBram(Bram#(addr_t, data_t))
    provisos(Bounded#(addr_t), Bits#(addr_t, addr_sz), Bits#(data_t, data_sz), Literal#(addr_t));

    no_reset;

    parameter addr_width = valueOf(addr_sz);
    parameter data_width = valueOf(data_sz);
    parameter hi = valueOf(TSub#(TExp#(addr_sz),1));
    parameter lo = 0;

    method readReq(read_addr) enable(read_en);
    method read_data readResp();
    method write(write_addr, write_data) enable(write_en);
    method zero() enable(RST);

    schedule readReq C readReq;
    schedule write C write;
    schedule readResp CF (readReq, readResp, write);
    schedule readReq CF write;
endmodule

module mkBramInstance(Bram#(Bit#(4), Bit#(5)));
    Bram#(Bit#(4), Bit#(5)) bram <- mkBram();
    return bram;
endmodule

module mkBramTest(Empty);
    Bram#(Bit#(4), Bit#(5)) bram <- mkBram();
    Reg#(Bit#(5)) val <- mkReg(11);

    rule all(True);
        bram.readReq(4);
        bram.write(9, val);
        $display("both");
    endrule
endmodule
