import Interfaces::*;

import Types::*;
import FuncUnit::*;
import FIFO::*;
import FIFOF::*;
import GetPut::*; 
import PLBMasterBackup::*;
import BRAMFeeder::*;
import Connectable::*;
import BRAMInitiatorWires::*;

(* synthesize *)
module mkMultiplierBackupTop(MultiplierBackup);
    FuncUnit fu <- mkfunctionalunit();

    PLBMasterBackup plb     <- mkPLBMasterBackup();
    Feeder feeder     <- mkBRAMFeeder();

    Reg#(Bit#(32)) counter <- mkReg(0);

    FIFOF#(Data) dataInFIFO  <- mkFIFOF();
    FIFOF#(Data) dataOutFIFO <- mkFIFOF();
    FIFOF#(Instruction) instFIFO <- mkFIFOF();

    let inst = instFIFO.first();
    let fuOp = !(inst.op == SetRowSize || inst.op == LoadAddr || inst.op == StoreAddr);

    rule processInst(True);
        let instruction = feeder.ppcInstructionOutput.get();
        instFIFO.enq(inst);
    endrule

    mkConnection(fu.putData, plb.wordOutput);
    mkConnection(fu.getData, plb.wordInput);

    rule processFUOp(fuOp);
        fu.putInst.put(inst.op);
        instFIFO.deq();
    endrule

    rule processSetRowSize(inst.op == SetRowSize);
        plb.plbMasterCommandInput.put(tagged RowSize truncate(inst.addr));
        instFIFO.deq();
    endrule

    rule processLoadAddr(inst.op == LoadAddr);
        plb.plbMasterCommandInput.put(tagged LoadPage inst.addr);
        instFIFO.deq();
    endrule

    rule processStoreAddr(inst.op == StoreAddr);
        plb.plbMasterCommandInput.put(tagged StorePage inst.addr);
        instFIFO.deq();
    endrule

    rule processStoreReturn(True);
        let store <- fu.getStore.get();
        counter <= counter + 1;
        feeder.ppcMessageInput.put(counter);
    endrule

    interface plbBRAMWires     = plb.plbBRAMWires;
    interface bramInitiatorWires = feeder.bramInitiatorWires;
endmodule
