module bram_fu
(
    CLK,
    CLK_GATE,
    RST,
    write_en,
    write_addr,
    write_data,
    read_en,
    read_addr,
    read_data
);
    parameter addr_width = 9;
    parameter data_width = 1024;
    parameter lo = 0;
    parameter hi = 511;

    input CLK;
    input CLK_GATE;
    input RST;
    input write_en;
    input  [addr_width-1:0] write_addr;
    input  [data_width-1:0] write_data;
    input read_en;
    input  [addr_width-1:0] read_addr;
    output reg [data_width-1:0] read_data;

    wire en = write_en | read_en;
    
    //synthesis attribute ram_extract of RAM is yes;
    //synthesis attribute ram_style of RAM is block;
    reg [data_width-1:0] ram[hi:lo];

    integer x;

    always@(posedge CLK)
    begin
        if(RST)
        begin
            // synopsys translate_off
            for (x = lo; x <= hi; x = x + 1)
            begin
                ram[x] <= 0;
            end
            // synopsys translate_on
            read_data <= 0;
        end
        else
        if(en)
        begin
            read_data <= ram[read_addr];
            if(write_en)
                ram[write_addr] <= write_data;
        end
    end
endmodule
