module multiplyAdder
(
    CLK,
    CLK_GATE,
    x,
    y,
    accum,
    is_load,
    mulAdd,
    enable
);

    input CLK;
    input CLK_GATE;
    input  [31:0] x;
    input  [31:0] y;
    input         is_load;
    input  [31:0] accum;
    output [31:0] mulAdd;

    input enable;

    wire [17:0] a = {{2{x[31]}}, x[31:16]};
    wire [17:0] b = {{2{x[15]}}, x[15:0]};
    wire [17:0] c = {{2{y[31]}}, y[31:16]};
    wire [17:0] d = {{2{y[15]}}, y[15:0]};

    reg [31:0] ac;
    reg [31:0] bd;
    reg [31:0] ad;
    reg [31:0] bc;

    wire [31:0] ac_wire = a*c;
    wire [31:0] bd_wire = b*d;
    wire [31:0] ad_wire = a*d;
    wire [31:0] bc_wire = b*c;

    always@(posedge CLK)
    begin
        if(enable)
        begin
            ac <= {{14{ac_wire[31]}}, ac_wire[31:14]};
            bd <= {{14{bd_wire[31]}}, bd_wire[31:14]};
            ad <= {{14{ad_wire[31]}}, ad_wire[31:14]};
            bc <= {{14{bc_wire[31]}}, bc_wire[31:14]};
        end
    end

    reg [31:0] accum_reg1;

    always@(posedge CLK)
    begin
        if(enable)
        begin
            if(is_load)
                accum_reg1 <= accum;
            else
                accum_reg1 <= mulAdd;
        end
    end

    wire [15:0] r_part = ac[15:0]-bd[15:0];
    wire [15:0] i_part = ad[15:0]+bc[15:0];

    wire [15:0] real_part = r_part+accum_reg1[31:16];
    wire [15:0] imag_part = i_part+accum_reg1[15:0];

    assign mulAdd = {real_part, imag_part};

endmodule
