import ISA::*;
import GetPut::*;
import FIFOF::*;

interface Decoder;
    interface Put#(Inst) put; 
endinterface

module mkDecoder(Decoder);
    FIFOF#(Inst) instFIFO <- mkFIFOF();

    LoadUnit   loadUnit <- mkLoadUnit();
    StoreUnit storeUnit <- mkStoreUnit();
    FunctionalUnit#(8) functionalUnit <- mkFunctionalUnit8();
    ForwardUnit       forwardUnit <- mkForwardUnit();

    Reg#(Bool) busy <- mkReg(False);

    let inst = instFIFO.first();

    rule loadReq(!busy && inst matches Tagged Load .memMesg);
        loadUnit.put.put(memMesg);
        busy <= True;
    endrule

    rule loadResp(busy && inst matches Tagged Load .memMesg);
        let val <- loadUnit.get.get();
        instFIFO.deq();
        busy <= False;
    endrule

    rule storeReq(!busy && inst matches Tagged Store .memMesg);
        storeUnit.put.put(memMesg);
        busy <= True;
    endrule

    rule storeResp(busy && inst matches Tagged Store .memMesg);
        let val <- storeUnit.get.get();
        instFIFO.deq();
        busy <= False;
    endrule

    rule setRowSize(!busy && inst matches Tagged SetRowSize .rowSize);
        loadUnit.setRowSize(rowSize);
        storeUnit.setRowSize(rowSize);
        instFIFO.deq();
    endrule

    rule forwardReq(!busy && inst matches Tagged Forward .forward);
        forwardUnit.put.put(forward);
        busy <= True;
    endrule

    rule forwardResp(busy && inst matches Tagged Forward .forward);
        let val <- forwardUnit.get.get();
        instFIFO.deq();
        busy <= False;
    endrule

    rule restReq(!busy);
        functionalUnit.put.put(inst);
        busy <= True;
    endrule

    rule restResp(busy);
        let val <- funcationalUnit.get.get();
        instFIFO.deq();
        busy <= False;
    endrule 
endmodule
